import functools

from typing import Union


class BaseExpression(object):
    def __new__(cls, *args, **kwargs):
        if cls is BaseExpression:
            raise TypeError("'BaseExpression' class may not be instantiated directly")
        return object.__new__(cls)

    def evaluate(self, **kwargs):
        raise NotImplementedError

    @property
    def variables(self) -> set:
        raise NotImplementedError

    def __add__(self, operand: "BaseExpression"):
        return AdditionExpression(self, operand)

    def __radd__(self, operand: "BaseExpression"):
        return AdditionExpression(operand, self)

    def __sub__(self, operand: "BaseExpression"):
        return SubtractionExpression(self, operand)

    def __rsub__(self, operand: "BaseExpression"):
        return SubtractionExpression(operand, self)

    def __mul__(self, operand: "BaseExpression"):
        return MultiplicationExpression(self, operand)

    def __rmul__(self, operand: "BaseExpression"):
        return MultiplicationExpression(operand, self)

    def __truediv__(self, operand: "BaseExpression"):
        return DivisionExpression(self, operand)

    def __rtruediv__(self, operand: "BaseExpression"):
        return DivisionExpression(operand, self)

    def __pow__(self, operand: "BaseExpression"):
        return ExponentiationExpression(self, operand)

    def __rpow__(self, operand: "BaseExpression"):
        return ExponentiationExpression(operand, self)

    def __neg__(self):
        return NegativeExpression(self)


class BinaryExpression(BaseExpression):
    _operator: str

    def __new__(cls, *args, **kwargs):
        if cls is BaseExpression:
            raise TypeError("'BinaryExpression' class may not be instantiated directly")
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, operand1: BaseExpression, operand2: BaseExpression) -> None:
        self._operand1 = operand1
        self._operand2 = operand2

    def evaluate(self, **kwargs):
        if type(self._operand1) in [int, float, complex]:
            left_hand_side = self._operand1
        else:
            left_hand_side = self._operand1.evaluate(**kwargs)

        if type(self._operand2) in [int, float, complex]:
            right_hand_side = self._operand2
        else:
            right_hand_side = self._operand2.evaluate(**kwargs)

        return self._apply(left_hand_side, right_hand_side)

    @property
    def variables(self) -> set:
        return self._operand1.variables.union(self._operand2.variables)

    def _apply(self, left_hand_side, right_hand_side):
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._operand1)}, {repr(self._operand2)})"

    def __str__(self) -> str:
        return f"{str(self._operand1)} {self._operator} {str(self._operand2)}"


class UnaryExpression(BaseExpression):
    _operator: str

    def __new__(cls, *args, **kwargs):
        if cls is BaseExpression:
            raise TypeError("'UnaryExpression' class may not be instantiated directly")
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, operand: BaseExpression) -> None:
        self._operand = operand

    def evaluate(self, **kwargs):
        if type(self._operand) in [int, float, complex]:
            value = self._operand
        else:
            value = self._operand.evaluate(**kwargs)

        return self._apply(value)

    @property
    def variables(self) -> set:
        return self._operand.variables

    def _apply(self, value):
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._operand)})"

    def __str__(self) -> str:
        return f"{self._operator}{str(self._operand)}"


class Number(BaseExpression):
    def __init__(self, value: str) -> None:
        try:
            self._value = int(value)
        except ValueError:
            try:
                self._value = float(value)
            except ValueError:
                self._value = complex(value)

    def evaluate(self, **kwargs):
        return self._value

    @property
    def variables(self) -> set:
        return set()

    def __repr__(self):
        return f"Number('{repr(self._value)}')"

    def __str__(self):
        return str(self._value)


class Variable(BaseExpression):
    def __init__(self, name: str) -> None:
        self._name = name

    def evaluate(self, **kwargs):
        if self._name not in kwargs:
            raise ValueError(f"Variable `{self._name}` does not have a value!")

        return kwargs[self._name]

    @property
    def variables(self) -> set:
        return {self._name}

    def __repr__(self) -> str:
        return f"Variable('{self._name}')"

    def __str__(self) -> str:
        return self._name


class Function(BaseExpression):
    supported_functions = {}

    def __init__(self, name: str, argument: "FunctionArgument") -> None:
        self._name = name
        self._argument = argument

    def evaluate(self, **kwargs):
        if self._name not in Function.supported_functions.keys():
            raise ValueError(f"Function `{self._name}` has not been defined!")

        return Function.supported_functions[self._name](*self._argument.evaluate(**kwargs))

    @property
    def variables(self) -> set:
        return self._argument.variables

    @staticmethod
    def register(name: str, runner: callable):
        Function.supported_functions[name] = runner

    def __repr__(self) -> str:
        return f"Function('{self._name}', {repr(self._argument)})"

    def __str__(self) -> str:
        return f"{self._name}({str(self._argument)})"


register_function = Function.register


class FunctionArgument(object):
    def __init__(self, *expression: BaseExpression) -> None:
        self._expressions = [*expression]

    def evaluate(self, **kwargs):
        return self._map(lambda e: e.evaluate(**kwargs))

    @property
    def variables(self) -> set:
        return functools.reduce(lambda a, b: a.union(b), self._map(lambda e: e.variables))

    def _map(self, fn):
        return [fn(expression) for expression in self._expressions]

    def __add__(self, op: Union[BaseExpression, 'FunctionArgument']) -> 'FunctionArgument':
        if type(op) != FunctionArgument and isinstance(op, BaseExpression) is False:
            raise NotImplementedError()

        if type(op) == FunctionArgument:
            return FunctionArgument(*self._expressions, *op._expressions)

        return FunctionArgument(*self._expressions, op)

    def __radd__(self, op: Union[BaseExpression, 'FunctionArgument']) -> 'FunctionArgument':
        if type(op) not in [BaseExpression, FunctionArgument]:
            raise NotImplementedError()

        if type(op) == FunctionArgument:
            return FunctionArgument(*op._expressions, *self._expressions)

        return FunctionArgument(op, *self._expressions)

    def __repr__(self) -> str:
        return f"FunctionArgument({', '.join(self._map(repr))})"

    def __str__(self) -> str:
        return f"{', '.join(self._map(str))}"


class AdditionExpression(BinaryExpression):
    _operator = "+"

    def __init__(self, operand1: BaseExpression, operand2: BaseExpression) -> None:
        super().__init__(operand1, operand2)

    def _apply(self, left_hand_side, right_hand_side):
        return left_hand_side + right_hand_side


class SubtractionExpression(BinaryExpression):
    _operator = "-"

    def __init__(self, operand1: BaseExpression, operand2: BaseExpression) -> None:
        super().__init__(operand1, operand2)

    def _apply(self, left_hand_side, right_hand_side):
        return left_hand_side - right_hand_side


class MultiplicationExpression(BinaryExpression):
    _operator = "*"

    def __init__(self, operand1: BaseExpression, operand2: BaseExpression) -> None:
        super().__init__(operand1, operand2)

    def _apply(self, left_hand_side, right_hand_side):
        return left_hand_side * right_hand_side


class DivisionExpression(BinaryExpression):
    _operator = "/"

    def __init__(self, operand1: BaseExpression, operand2: BaseExpression) -> None:
        super().__init__(operand1, operand2)

    def _apply(self, left_hand_side, right_hand_side):
        return left_hand_side / right_hand_side


class ExponentiationExpression(BinaryExpression):
    _operator = "^"

    def __init__(self, operand1: BaseExpression, operand2: BaseExpression) -> None:
        super().__init__(operand1, operand2)

    def _apply(self, left_hand_side, right_hand_side):
        return left_hand_side ** right_hand_side


class NegativeExpression(UnaryExpression):
    _operator = "-"

    def __init__(self, operand: BaseExpression) -> None:
        super().__init__(operand)

    def _apply(self, value):
        return -value



class KharazmiBaseError(Exception):
    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class ParseError(KharazmiBaseError):
    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class EvaluationError(KharazmiBaseError):
    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class LexError(ParseError):
    def __init__(self, message: str = "") -> None:
        super().__init__(message)



from sly import Lexer
from sly.lex import Token

from .exceptions import LexError


class EquationLexer(Lexer):
    tokens = {
        NUMBER,
        IDENTIFIER,
        PLUS,
        MINUS,
        TIMES,
        DIVIDE,
        POWER,
    }

    literals = ["(", ")", ","]
    ignore = " \t"

    IDENTIFIER = r"[a-zA-Z_][a-zA-Z_0-9]*"
    NUMBER = r"\d+(\.\d*)?((\+|\-)\d+(\.\d*)?j)?"
    PLUS = r'\+'
    MINUS = r'-'
    TIMES = r'\*'
    DIVIDE = r'/'
    POWER = r'\^'

    def error(self, token: Token):
        raise LexError(f"Invalid token '{token.value[0]}'")



import math

from .models import Function

functions = {
    "abs": abs,
    "max": max,
    "min": min,
    "round": round,
    'acos': math.acos,
    'acosh': math.acosh,
    'asin': math.asin,
    'asinh': math.asinh,
    'atan': math.atan,
    'atan2': math.atan2,
    'atanh': math.atanh,
    'ceil': math.ceil,
    'copysign': math.copysign,
    'cos': math.cos,
    'cosh': math.cosh,
    'degrees': math.degrees,
    'erf': math.erf,
    'erfc': math.erfc,
    'exp': math.exp,
    'expm1': math.expm1,
    'fabs': math.fabs,
    'factorial': math.factorial,
    'floor': math.floor,
    'fmod': math.fmod,
    'frexp': math.frexp,
    'fsum': math.fsum,
    'gamma': math.gamma,
    'gcd': math.gcd,
    'hypot': math.hypot,
    'isclose': math.isclose,
    'isfinite': math.isfinite,
    'isinf': math.isinf,
    'isnan': math.isnan,
    'ldexp': math.ldexp,
    'lgamma': math.lgamma,
    'log': math.log,
    'log1p': math.log1p,
    'log10': math.log10,
    'log2': math.log2,
    'modf': math.modf,
    'pow': math.pow,
    'radians': math.radians,
    'remainder': math.remainder,
    'sin': math.sin,
    'sinh': math.sinh,
    'sqrt': math.sqrt,
    'tan': math.tan,
    'tanh': math.tanh,
    'trunc': math.trunc,
}


def activate_builtin_math():
    for name, fn in functions.items():
        Function.register(name, fn)



#!/usr/bin/env python

import os
import subprocess


from setuptools import setup, find_packages


def run_setup():
    requirements = [
        "sly>=0.4",
    ]

    python_requirement = ">=3.6"

    with open('README.md', 'r') as f:
        long_description = f.read()

    setup(
        name='kharazmi',
        version=git_version(),

        author='Mahdi Mohaveri',
        author_email='mmohaveri@gmail.com',

        package_dir={'': 'src'},
        packages=find_packages('src'),
        package_data={
            'kharazmi': ['py.typed'],
        },
        include_package_data=True,

        license='MIT License',

        description='An equation parser/calculator module for python.',
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/mmohaveri/kharazmi",

        install_requires=requirements,
        python_requires=python_requirement,

        zip_safe=True,
        keywords=[
            'development',
            'expression',
            'algebra',
            'equation'
        ],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Topic :: Software Development',
            'Topic :: Utilities',
            'Typing :: Typed'
        ],
    )


def git_version():
    git_commit_cmd = ['git', 'rev-parse', '--short', 'HEAD']
    git_version_cmd = ['git', 'describe', '--abbrev=1', '--tags']

    return _run_cmd(git_version_cmd) or _run_cmd(git_commit_cmd) or "Unknown"


def _run_cmd(cmd):
    cwd = os.path.abspath(os.path.dirname(__file__))

    env = {key: os.environ.get(key) for key in [
        'SYSTEMROOT', 'PATH', 'HOME'] if os.environ.get(key) is not None}
    env['LANGUAGE'] = 'C'
    env['LANG'] = 'C'
    env['LC_ALL'] = 'C'

    child = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env, cwd=cwd)
    output, err = child.communicate()

    if err is not None or child.returncode != 0:
        return None

    return output.strip().decode('ascii')


if __name__ == "__main__":
    run_setup()



from typing import Union
from .lexer import EquationLexer
from .parser import EquationParser
from . import builtin_math


def main():
    parser = EquationParser()

    while True:
        cmd = input("> ")

        if check_for_exit(cmd):
            break

        expression = parser.parse(cmd)
        print(f"You've entered: {str(expression)}")
        print(f"You can do it in code using: {repr(expression)}")

        kwargs = {}

        for var in expression.variables:
            kwargs[var] = number_input(f"{var} = ")

        print(f"Result is: {expression.evaluate(**kwargs)}")


def number_input(message: str) -> Union[int, float, complex]:
    val = input(message)

    try:
        return int(val)
    except ValueError:
        try:
            return float(val)
        except ValueError:
            return complex(val)


def check_for_exit(cmd: str) -> bool:
    return cmd.startswith("exit")


if __name__ == "__main__":
    main()



from sly import Parser

from .exceptions import ParseError
from .models import Function, Number, Variable, FunctionArgument
from .lexer import EquationLexer


class EquationParser(Parser):
    """
    EquationParser implements a CFG parser for the following grammar:

    expression : expression PLUS expression
               | expression MINUS expression
               | expression TIMES expression
               | expression DIVIDE expression
               | expression POWER expression
               | ( expression )
               | MINUS expression
               | NUMBER
               | IDENTIFIER
               | IDENTIFIER ( argument )

    argument   : expression
               | argument , expression
    """

    def __init__(self):
        self.lexer = EquationLexer()

    def parse(self, inp: str):
        return super().parse(self.lexer.tokenize(inp))

    tokens = EquationLexer.tokens

    precedence = (
        ("left", PLUS, MINUS),
        ("left", TIMES, DIVIDE),
        ("left", POWER),
        ('right', UMINUS),
    )

    start = 'expression'

    @_("expression PLUS expression")
    def expression(self, p):
        return p.expression0 + p.expression1

    @_("expression MINUS expression")
    def expression(self, p):
        return p.expression0 - p.expression1

    @_("expression TIMES expression")
    def expression(self, p):
        return p.expression0 * p.expression1

    @_("expression DIVIDE expression")
    def expression(self, p):
        return p.expression0 / p.expression1

    @_("expression POWER expression")
    def expression(self, p):
        return p.expression0 ** p.expression1

    @_("'(' expression ')'")
    def expression(self, p):
        return p.expression

    @_("MINUS expression %prec UMINUS")
    def expression(self, p):
        return -p.expression

    @_("NUMBER")
    def expression(self, p):
        return Number(p[0])

    @_("IDENTIFIER")
    def expression(self, p):
        return Variable(p[0])

    @_("IDENTIFIER '(' argument ')'")
    def expression(self, p):
        return Function(p[0], p[2])

    @_("expression")
    def argument(self, p):
        return FunctionArgument(p.expression)

    @_("argument ',' expression")
    def argument(self, p):
        return p.argument + p.expression

    def error(self, p):
        if p is None:
            raise ParseError(f"Incomplete expression.")

        raise ParseError(f"Invalid expression. Error occurred in position {p.index}")
